#encoding:utf-8
from os import path
import multiprocessing
import os

BASE_DIR = 'nlptooldemo'
RE = {
'pretrained_model_path' :os.path.join(BASE_DIR, '../../bert/pretrained_models/bert-base-chinese'),
'train_file' : os.path.join(BASE_DIR, 'dataset/REdataset/train_small.jsonl'),
'validation_file' : os.path.join(BASE_DIR, 'dataset/REdataset/val_small.jsonl'),
'output_dir' : os.path.join(BASE_DIR, 'dataset/REdataset/saved_models'),
'tagset_file' : os.path.join(BASE_DIR, 'dataset/REdataset/relation.txt'),
'max_len':128,
'ner_source': path.sep.join([BASE_DIR, 'output/result/predict_result.txt']),
'checkpoint':path.sep.join([BASE_DIR,'output/RE_checkpoint/re-best.pth']),
'result':path.sep.join([BASE_DIR,'output/re.txt']),
'dataset':os.path.join(BASE_DIR,'dataset')
}
configs = {
    'all_data_path': [path.sep.join([BASE_DIR,'dataset/raw/msra_source_train.txt']),path.sep.join([BASE_DIR,'dataset/raw/pku_cws_source.txt']),path.sep.join([BASE_DIR,'dataset/raw/msr_cws_source.txt']) ],  
    'pretrained_path':path.sep.join([BASE_DIR, 
                                            'output/checkpoints/roberta.zip']),# for word embedding
    'test_file_path': path.sep.join([BASE_DIR,'dataset/processed/test.json']),   
    'embedding_weight_path': path.sep.join([BASE_DIR, 
                                            'output/embedding/sgns300']), # for character embedding
    'embedding_dict_path': path.sep.join([BASE_DIR, 
                                            'output/embedding/sgns300']),# for word embedding
    'vocab_path': path.sep.join([BASE_DIR,'dataset/processed/vocab.pkl']), 
    'rev_vocab_path': path.sep.join([BASE_DIR,'dataset/processed/rev_vocab.pkl']), 

    'log_dir': path.sep.join([BASE_DIR, 'output/log']), 
    'writer_dir': path.sep.join([BASE_DIR, 'output/TSboard']),
    'figure_dir': path.sep.join([BASE_DIR, 'output/figure']),
    'checkpoint_dir': path.sep.join([BASE_DIR, 'output/checkpoints/bilstm_crf']),
    'embedding_dir': path.sep.join([BASE_DIR, 'output/embedding']),

    'valid_size': 0.1, 
    'min_freq': 1,   
    'max_length': 80,  
    'max_features': 100000, 
    'embedding_dim':300, 

    'batch_size': 32,  
    'epochs': 100,    
    'start_epoch': 1,
    'learning_rate': 3e-5, #for bert
    'weight_decay': 5e-4, 
    'n_gpus': [0], 
    'x_var':'source', 
    'y_var':'target', 
    'num_workers': multiprocessing.cpu_count(), 
    'seed': 2018, 
    'lr_patience': 5, 
    'mode': 'min',    
    'monitor': 'val_loss',  
    'early_patience': 10,   
    'save_best_only':True, 
    'best_model_name': '{arch}-best-re.pth', 
    'epoch_model_name': '{arch}-{epoch}-{val_loss}.pth', 
    'save_checkpoint_freq': 10,

    'multi-task':
    [

  {    'label_to_id': {
  
        "B_PER": 1,  
        "I_PER": 2,
        "B_LOC": 3, 
        "I_LOC": 4,
        "B_ORG": 5,  
        "I_ORG": 6,
        "B_T":7,
        "I_T":8,
        "O": 9 

  },
    'num_tag':10,
    'raw_train_path': path.sep.join([BASE_DIR,'dataset/raw/boson_ner_source.txt']),  
    'raw_target_path': path.sep.join([BASE_DIR,'dataset/raw/boson_ner_target.txt']), 
    'raw_test_path': path.sep.join([BASE_DIR,'dataset/raw/test.txt']),       
    'test_file_path': path.sep.join([BASE_DIR,'dataset/processed/test.json']),   
    'result_path': path.sep.join([BASE_DIR, 'output/result/predict_result.txt']),
    'x_var':'source', 
    'y_var':'target',
    'train_file_path': path.sep.join([BASE_DIR,'dataset/processed/train.json']), 
    'valid_file_path': path.sep.join([BASE_DIR,'dataset/processed/valid.json']),
    'name':"NER"
    }

    ],
    'seed':2018,
    'resume':False,
    'model':{'name':'bert_lstm'
    },
    
    'models': {
    'bert_lstm':{'hidden_size': 200,
                              'bi_tag': True,
                             'dropout_p':0.5,
                             'dropout_emb':0.0,
                             'num_layer': 1,
                             'use_cuda':True,
                             'crf':True},
    'roberta_lstm':{'hidden_size': 200,
                              'bi_tag': True,
                             'dropout_p':0.5,
                             'dropout_emb':0.0,
                             'num_layer': 1,
                             'use_cuda':True,
                             'crf':True},
    'lattice_lstm':{'hidden_size': 200,
                             'bi_tag': True,
                             'dropout_p':0.5,
                             'dropout_emb':0.0,
                             'num_layer': 1,
                             'use_cuda':True,
                             'crf':True},
    'cnn_crf':{'hidden_size': 200,
                             'bi_tag': True,
                             'dropout_p':0.5,
                             'dropout_emb':0.0,
                             'num_layer': 1,
                             'use_cuda':True,
                             'crf':True},
    'bilstm':{'hidden_size': 200,
                             'bi_tag': True,
                             'dropout_p':0.5,
                             'dropout_emb':0.0,
                             'num_layer': 1,
                             'use_cuda':True,
                             'crf':False},
    'bilstm_crf':{'hidden_size': 200,
                         'bi_tag': True,
                             'dropout_p':0.5,
                             'dropout_emb':0.0,
                             'num_layer': 1,
                             'use_cuda':True,
                             'crf':True}
              }
}



import collections
class TrieNode:
    # Initialize your data structure here.
    def __init__(self):
        self.children = collections.defaultdict(TrieNode)
        self.is_word = False
        self.word_num = -1

class Trie:
    def __init__(self):
        self.root = TrieNode()

    def insert(self, word, num):
        
        current = self.root
        for letter in word:
            if (letter == 0):
                break
            current = current.children[letter]
        current.is_word = True
        current.word_num = num

    def search(self, word):
        current = self.root
        for letter in word:
            if (letter == 0 ):
                break
            current = current.children.get(letter)

            if current is None:
                return False, -1
        return current.is_word, current.word_num

    def startsWith(self, prefix):
        current = self.root
        for letter in prefix:
            current = current.children.get(letter)
            if current is None:
                return False
        return True


    def enumerateMatch(self, word, space="_", backward=False):
        matched = []
        matched_lens = []
        ## while len(word) > 1 does not keep character itself, while word keed character itself
        while len(word) > 1:

            wd, num = self.search(word)
            if wd:
                matched.append(num)
                matched_lens.append(len(word))
                
            del word[-1]
        return [matched,matched_lens]




#encoding:utf-8
import os
import random
import torch
import json
import pickle
import numpy as np

# 设置seed环境
def seed_everything(seed = 1029):
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True

class AverageMeter(object):
    '''
    computes and stores the average and current value
    '''
    def __init__(self):
        self.reset()
    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self,val,n = 1):
        self.val  = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


def ensure_dir(path):
    if not os.path.exists(path):
        os.makedirs(path)

def json_write(data,filename):
    with open(filename,'w') as f:
        json.dump(data,f)

def json_read(filename):
    with open(filename,'r') as f:
        return json.load(f)

def text_write(file,data,x_var,y_var):
    with open(file, 'w') as fw:
        for sent, label in data:
            sent = ' '.join([str(s) for s in sent])
            label = ' '.join([str(x) for x in label])
            df = {x_var:sent,
                  y_var:label}
            encode_json = json.dumps(df)
            print(encode_json,file = fw)

def pkl_read(filename):
    with open(filename,'rb') as f:
        return pickle.load(f)

def pkl_write(filename,data):
    with open(filename, 'wb') as f:
        pickle.dump(data, f)





#encoding:Utf-8
class ProgressBar():
    def __init__(self,n_batch,
                 eval_name = 'acc',
                 loss_name = 'loss',
                 width=30
                 ):
        self.width = width
        self.loss_name = loss_name
        self.eval_name = eval_name
        self.n_batch = n_batch
        self.use = 'on_batch_end'

    def step(self,batch_idx,loss,acc,use_time,f1):
        recv_per = int(100 * (batch_idx + 1) / self.n_batch)
        if recv_per >= 100:
            recv_per = 100
        # 只显示train数据结果
        show_bar = ('[%%-%ds]' % self.width) % (int(self.width * recv_per / 100) * ">")
        show_str = '\r[training] %d/%d %s -%.1fs/step- %s: %.4f- %s: %.4f - f1: %.4f'
        print(show_str % (
            batch_idx+1,self.n_batch,show_bar, use_time,self.loss_name, loss,self.eval_name, acc,f1),end='')





import argparse
import torch
import numpy as np
import warnings
import os
from torch import optim
from itertools import chain
from .test.predicter import Predicter
from .test.predict_utils import test_write
from .train.metrics import F1_score
from .train.trainer import Trainer
from .train.train_utils import restore_checkpoint,model_device
from torch.utils.data.distributed import DistributedSampler
from torch.utils.data import DataLoader

from .io.dataset import MultiDataset, REDataset, get_idx2tag
from .io.data_transformer import DataTransformer
from .io.downloader import Downloader

from .utils.logginger import init_logger
from .utils.utils import seed_everything
from .config.new_config import configs as config

from .config.new_config import RE as hparams


from .model.nn.cnn import CNN
from .model.nn.bilstm_crf import BiLSTM
from .model.nn.latticelstm_crf import Lattice
from .model.nn.bert import BERT_LSTM
from .model.nn.roberta import RoBERTa_LSTM
from .model.nn.relation import SentenceRE
from .callback.modelcheckpoint import ModelCheckpoint
from tqdm import tqdm

import torch.distributed as dist
import zipfile
import abc

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
warnings.filterwarnings("ignore")

def synchronize():

    if not dist.is_available():
        return
    if not dist.is_initialized():
        return
    world_size = dist.get_world_size()
    if world_size == 1:
        return
    dist.barrier()

class Component(abc.ABC):
    def __init__(self, config, local_rank):	
        super().__init__()
        self.config = config
        self.local_rank = local_rank
        self.num_gpus = int(os.environ["WORLD_SIZE"]) if "WORLD_SIZE" in os.environ else 1
        self.is_distributed = (self.num_gpus>1)
        if self.is_distributed:
            torch.cuda.set_device(self.local_rank)
            self.device = 'cuda:%d'%(self.local_rank)
            torch.distributed.init_process_group(
            backend="nccl", init_method="env://"
        )
            synchronize()
        else:
          self.device = 'cuda:%d' % config['n_gpus'][0] if len(config['n_gpus']) else 'cpu'
          if (len(config['n_gpus'])):
            torch.cuda.set_device(config['n_gpus'][0])
    @abc.abstractmethod
    def train():
        pass
    @abc.abstractmethod
    def predict():
        pass
class RelationExtractor(Component):
    def __init__(self, config, local_rank):
        super().__init__(config, local_rank)
        if self.local_rank == 0:
            self.dl = Downloader()
            if (not os.path.exists(hparams['dataset']+'/REdataset')):
                self.dl.download("REdataset",hparams['dataset']+'/REdataset.zip')
                f = zipfile.ZipFile(hparams['dataset']+'/REdataset.zip','r')
                for file in f.namelist():
                    print("extracting:",str(file),'...')
                    f.extract(file,hparams['dataset']+'/REdataset')
                print("extracted.")
   #     self.logger = init_logger(log_name=self.model_config['name'], log_dir=config['log_dir'])
    def train(self):
        self.criterion = torch.nn.CrossEntropyLoss().cuda()
        RE_train_dataset =  REDataset(hparams['train_file'], tagset_path=hparams['tagset_file'],
                                      pretrained_model_path=None,
                                      max_len=hparams['max_len'])
        RE_val_dataset =  REDataset(hparams['validation_file'], tagset_path=hparams['tagset_file'],
                                      pretrained_model_path=None,
                                      max_len=hparams['max_len'])
        self.REModel = SentenceRE(embedding_dim=768,dropout=0.1,tagset_size=49)
        self.REoptimizer = torch.optim.Adam(self.REModel.parameters(), lr=1e-5, weight_decay=0)
        self.RE_train_loader = DataLoader(RE_train_dataset, batch_size=32, shuffle=False,sampler=DistributedSampler(RE_train_dataset,shuffle=False))
        self.RE_val_loader = DataLoader(RE_val_dataset, batch_size=32, shuffle=False)              
        self.REModel = self.REModel.cuda()
        if self.is_distributed:
            self.REModel = torch.nn.parallel.DistributedDataParallel(
            self.REModel, device_ids=[self.local_rank], output_device=self.local_rank,
        find_unused_parameters=True
            )  
        else:
        	pass

        if (self.REModel):
            epoch_offset = 0
            best_val = float("inf")
            for epoch in range(epoch_offset, 20):
                if self.local_rank==0:
                    print("Epoch: {}".format(epoch))
                self.REModel.train()
                idx=0
                for i_batch, sample_batched in enumerate(tqdm(self.RE_train_loader, desc='Training')):
                    idx+=32

                    token_ids = sample_batched['token_ids'].cuda()
                    token_type_ids = sample_batched['token_type_ids'].cuda()
                    attention_mask = sample_batched['attention_mask'].cuda()
                    e1_mask = sample_batched['e1_mask'].cuda()
                    e2_mask = sample_batched['e2_mask'].cuda()
                    tag_ids = sample_batched['tag_id'].cuda()
                    self.REModel.zero_grad()
                    logits = self.REModel(token_ids, token_type_ids, attention_mask, e1_mask, e2_mask)
                    pred_tag_ids = logits.argmax(1)
                    ids = pred_tag_ids.cpu().numpy().tolist()


                    loss = self.criterion(logits, tag_ids)
                    loss.backward()
                    self.REoptimizer.step()
                    self.REoptimizer.zero_grad()
                if (self.local_rank==0):
                    with torch.no_grad():
                        val_loss = float(0.0)
                        for i_batch, sample_batched in enumerate(tqdm(self.RE_val_loader)):
                            token_ids = sample_batched['token_ids'].cuda()
                            token_type_ids = sample_batched['token_type_ids'].cuda()
                            attention_mask = sample_batched['attention_mask'].cuda()
                            e1_mask = sample_batched['e1_mask'].cuda()
                            e2_mask = sample_batched['e2_mask'].cuda()
                            tag_ids = sample_batched['tag_id'].cuda()
                            logits = self.REModel(token_ids, token_type_ids, attention_mask, e1_mask, e2_mask)
                            pred_tag_ids = logits.argmax(1)
                            ids = pred_tag_ids.cpu().numpy().tolist()


                            val_loss += self.criterion(logits, tag_ids)
                        if (float(val_loss) < float(best_val)):
                            if (self.local_rank==0):
                                print("val_loss:",float(best_val),'->',float(val_loss))
                            best_val = val_loss
                            if self.is_distributed:
                                state = { 'state_dict': self.REModel.module.state_dict(),
                                    'optimizer': self.REoptimizer.state_dict()}
                            else:
                                state = { 'state_dict': self.REModel.state_dict(),
                                    'optimizer': self.REoptimizer.state_dict()}    	
                            torch.save(state,hparams['checkpoint'])
                dist.barrier()
    def predict(self):
        RE_test_dataset =  REDataset(hparams['ner_source'], tagset_path=hparams['tagset_file'],
                                      pretrained_model_path=None,
                                      max_len=hparams['max_len'],ner2re = True)
        etts = RE_test_dataset.etts



        if self.local_rank==0:
            self.REModel = SentenceRE(embedding_dim=768,dropout=0.1,tagset_size=49)
            self.REoptimizer = torch.optim.Adam(self.REModel.parameters(), lr=1e-5, weight_decay=0)
            self.RE_test_loader = DataLoader(RE_test_dataset, batch_size=32, shuffle=False)
            self.REModel = self.REModel.cuda()
            checkpoint = torch.load(hparams['checkpoint'])
            self.REModel.load_state_dict(checkpoint['state_dict'])
            self.REoptimizer.load_state_dict(checkpoint['optimizer'])
            epoch_offset = 0
            idx2tag =  get_idx2tag(hparams['tagset_file'])
            with torch.no_grad():
                output_file = open(hparams['result'],'w')
                idx=0
                cur_sent = ''
                info = {'sent':'','info':[]}
                for i_batch, sample_batched in enumerate(tqdm(self.RE_test_loader, desc='Testing')):
                    token_ids = sample_batched['token_ids'].cuda()
                    token_type_ids = sample_batched['token_type_ids'].cuda()
                    attention_mask = sample_batched['attention_mask'].cuda()
                    e1_mask = sample_batched['e1_mask'].cuda()
                    e2_mask = sample_batched['e2_mask'].cuda()
                    tag_ids = sample_batched['tag_id']
                    logits = self.REModel(token_ids, token_type_ids, attention_mask, e1_mask, e2_mask)
                    pred_tag_ids = logits.argmax(1)
                    ids = pred_tag_ids.cpu().numpy().tolist()
                    for rel_id, source in zip(ids,tag_ids):
                        if (idx2tag[rel_id]=='unknown'):
                            idx+=1
                            continue
                        if (source == cur_sent):
                            info['info'].append(etts[idx]['head']['name']+'->'+etts[idx]['tail']['name']+':'+idx2tag[rel_id])
                        else:
                            cur_sent = source
                            print(str(info),file=output_file)
                            info = {'sent':source,'info':[etts[idx]['head']['name']+'->'+etts[idx]['tail']['name']+':'+idx2tag[rel_id]]}
                        idx+=1

class Labeler(Component):
    def __init__(self, config, local_rank):
        super().__init__(config, local_rank)
        self.model_config = config['model']
        self.logger = init_logger(log_name=self.model_config['name'], log_dir=config['log_dir'])
        seed_everything(seed = config['seed'])

        self.default_tk = True if ('bert' in self.model_config['name']) else False #tokenize


        if self.local_rank == 0:
            self.dl = Downloader()

            if ((not os.path.exists(config['embedding_dict_path']))):
                self.dl.download(url='embedding',file_path=config['embedding_dict_path'])
            if (('bert' in config['model']['name']) and (not os.path.exists(config['pretrained_path'][:-4]))):
                self.dl.download(url='roberta',file_path=config['pretrained_path'])
                f = zipfile.ZipFile(config['pretrained_path'],'r')
                for file in f.namelist():
                    print("extracting:",str(file),'...')
                    f.extract(file,config['pretrained_path'][:-4])
                print("extracted.")
        dist.barrier()

    def train(self):
        bs = self.config['batch_size']
        if ('lattice' in self.model_config['name']):
            bs = 1

        out_numbers = [] 
        self.data_transformer = DataTransformer(logger  = self.logger,
                                       is_train_mode = True,
                                       all_data_path = self.config['all_data_path'],
                                       vocab_path    = self.config['vocab_path'],
                                       rev_vocab_path    = self.config['rev_vocab_path'],
                                       max_features  = self.config['max_features'],
                                       label_to_id   = None,
                                       train_file    = None,
                                       valid_file      = None,
                                       valid_size      = self.config['valid_size'],
                                       min_freq      = self.config['min_freq'],
                                       seed          = self.config['seed'],
                                       default_token = self.default_tk)
        self.data_transformer.build_vocab()
        embedding_weight, words_embedding, gaz_tree = self.data_transformer.build_embedding_matrix(embedding_path = config['embedding_weight_path'], dict_path = config['embedding_dict_path'])
        train_files = []
        val_files = []
        for task in self.config['multi-task']: 
            try:
                out_numbers.append(task['num_tag']) 
                self.data_transformer.label_to_id = task['label_to_id'] 
                self.data_transformer.train_file = task['train_file_path']
                self.data_transformer.valid_file = task['valid_file_path']
                self.data_transformer.sentence2id(raw_data_path   = task['raw_train_path'],
                                 raw_target_path = task['raw_target_path'],
                                 x_var           = task['x_var'],
                                 y_var           = task['y_var'])

        
                task_train_file,valid_train_file = task['train_file_path'],task['valid_file_path']
                train_files.append(task_train_file)
                val_files.append(valid_train_file)

                x,y = task['x_var'],task['y_var']
            except:
              pass

        train_dataset = MultiDataset(num_threads = self.num_gpus,data_paths = train_files,x_var = x, y_var = y, skip_header = False, is_train_mode = True, batch_size = config['batch_size'],max_sentence_length = self.config['max_length'], default_token=self.default_tk, gaz=gaz_tree, vocab = self.data_transformer.vocab,rev_vocab = self.data_transformer.rev_vocab)
        train_loader = DataLoader(train_dataset, batch_size=self.config['batch_size'], shuffle=False, sampler=DistributedSampler(train_dataset,shuffle=False))
            
        val_dataset = MultiDataset(num_threads = self.num_gpus,data_paths = val_files,x_var = x, y_var = y, skip_header = False, is_train_mode = True,  batch_size = config['batch_size'],max_sentence_length = self.config['max_length'], default_token=self.default_tk, gaz=gaz_tree, vocab = self.data_transformer.vocab,rev_vocab = self.data_transformer.rev_vocab)
        val_loader = DataLoader(val_dataset, batch_size=self.config['batch_size'], shuffle=False, sampler=DistributedSampler(val_dataset,shuffle=False))

        arch = self.model_config['name']
        if (arch == 'cnn_crf' or arch == 'cnn'):
            self.model = CNN(out_numbers      = out_numbers,
                      embedding_dim    = self.config['embedding_dim'],
                      model_config     = self.config['models'][arch],
                      embedding_weight = embedding_weight,
                      vocab_size       = len(self.data_transformer.vocab),
                      device           = self.device)
        elif (arch =='bilstm' or arch == 'bilstm_crf'):
            self.model = BiLSTM(out_numbers      = out_numbers,
                      embedding_dim    = self.config['embedding_dim'],
                      model_config     = self.config['models'][arch],
                      embedding_weight = embedding_weight,
                      vocab_size       = len(self.data_transformer.vocab),
                      device           = self.device)

        elif (arch == 'lattice_lstm'):
            self.model = Lattice(out_numbers      = out_numbers,
                      embedding_dim    = self.config['embedding_dim'],
                      model_config     = self.config['models'][arch],
                      embedding_weight = embedding_weight,
                      vocab_size       = len(self.data_transformer.vocab),
                      dict_size = len(self.data_transformer.word_vocab),
                      pretrain_dict_embedding = words_embedding,
                      device           = self.device)
        elif (arch == 'bert_lstm'):
            self.model = BERT_LSTM(out_numbers      = out_numbers,
                      model_config     = self.config['models'][arch],
                      device           = self.device)
        elif (arch == 'roberta_lstm'):
            self.model = RoBERTa_LSTM(out_numbers      = out_numbers,
                      model_config     = self.config['models'][arch],
                      device           = self.device)


        self.optimizer = optim.Adam(params = self.model.parameters(),lr = self.config['learning_rate'],
                           weight_decay = self.config['weight_decay'])


        if self.is_distributed:
            self.model = self.model.cuda()
            self.model = torch.nn.parallel.DistributedDataParallel(
            self.model, device_ids=[self.local_rank], output_device=self.local_rank,
        find_unused_parameters=True
            )


        self.model_checkpoint = ModelCheckpoint(checkpoint_dir   = self.config['checkpoint_dir'],
                                       mode             = self.config['mode'],
                                       monitor          = self.config['monitor'],
                                       save_best_only   = self.config['save_best_only'],
                                       best_model_name  = self.config['best_model_name'],
                                       epoch_model_name = self.config['epoch_model_name'],
                                       arch             = arch,
                                       logger           = self.logger)



        self.scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(self.optimizer, mode='min', factor=0.7, patience=4,
                                                           verbose=True, min_lr=1e-9)

        self.trainer = Trainer(model            = self.model,
                      model_name       = arch,
                      train_data       = train_loader,
                      val_data         = val_loader,
                      optimizer        = self.optimizer,
                      epochs           = self.config['epochs'],
                      label_to_id      = self.config['multi-task'][0]["label_to_id"],
                      evaluate         = F1_score(num_classes=8),
                      logger           = self.logger,
                      model_checkpoint = self.model_checkpoint,
                      resume           = self.config['resume'],
                      lr_scheduler     = self.scheduler,
                      n_gpu            = self.config['n_gpus'],
                      num_tasks        = 2,
                      device           = self.device,
                      avg_batch_loss   = True,
                      distributed      = self.is_distributed)


      
        self.trainer.summary()

        self.trainer.train()

        if len(self.config['n_gpus']) > 0:
            torch.cuda.empty_cache()










    def predict(self):
        if (not self.local_rank == 0):
            return
        arch  = self.config['model']['name']
        self.logger = init_logger(log_name=arch, log_dir=config['log_dir'])
        seed_everything(seed = self.config['seed'])

        self.checkpoint_path = os.path.join(self.config['checkpoint_dir'].format(arch =arch),
                                    self.config['best_model_name'].format(arch = arch))

        word_dic = True if ('lattice' in self.model_config['name']) else False
        self.data_transformer = DataTransformer(
                     vocab_path    = self.config['vocab_path'],
                     rev_vocab_path    = self.config['rev_vocab_path'],
                      all_data_path = self.config['all_data_path'],
                     logger        = self.logger,
                     skip_header   = False,
                     is_train_mode = False,
                     default_token = self.default_tk,
                     seed          = self.config['seed'],
                     word_dic     = word_dic)
        self.data_transformer.build_vocab()
        embedding_weight, words_embedding, gaz_tree = self.data_transformer.build_embedding_matrix(embedding_path = self.config['embedding_weight_path'], dict_path = self.config['embedding_dict_path'])
        bs = self.config['batch_size']
        arch = self.model_config['name'] 
        if ('lattice' in arch):
            bs = 1
        test_files = []

        out_numbers = [] 
        for task in self.config['multi-task']: 
            try:
                out_numbers.append(task['num_tag']) 
                self.data_transformer.label_to_id = task['label_to_id'] 
                self.data_transformer.test_file = task['test_file_path']
                self.data_transformer.sentence2id(raw_data_path   = task['raw_test_path'],
                                 raw_target_path = None,
                                 x_var           = task['x_var'],
                                 y_var           = task['y_var'])

        
                task_test_file = task['test_file_path']
                test_files.append(task_test_file)


                x,y = task['x_var'],task['y_var']
            except:
              pass

        test_dataset = MultiDataset(num_threads = self.num_gpus,data_paths = test_files,x_var = x, y_var = y, skip_header = False, is_train_mode = True, batch_size = config['batch_size'],max_sentence_length = self.config['max_length'], default_token=self.default_tk, gaz=gaz_tree, vocab = self.data_transformer.vocab,rev_vocab = self.data_transformer.rev_vocab)
        test_loader = DataLoader(test_dataset, batch_size=self.config['batch_size'], shuffle=False)

        if (arch == 'cnn_crf' or arch == 'cnn'):
            self.model = CNN(out_numbers      = out_numbers,
                      embedding_dim    = self.config['embedding_dim'],
                      model_config     = self.config['models'][arch],
                      embedding_weight = embedding_weight,
                      vocab_size       = len(self.data_transformer.vocab),
                      device           = self.device)
        elif (arch =='bilstm' or arch == 'bilstm_crf'):
            self.model = BiLSTM(out_numbers      = out_numbers,
                      embedding_dim    = self.config['embedding_dim'],
                      model_config     = self.config['models'][arch],
                      embedding_weight = embedding_weight,
                      vocab_size       = len(self.data_transformer.vocab),
                      device           = self.device)

        elif (arch == 'lattice_lstm'):
            self.model = Lattice(out_numbers      = out_numbers,
                      embedding_dim    = self.config['embedding_dim'],
                      model_config     = self.config['models'][arch],
                      embedding_weight = embedding_weight,
                      vocab_size       = len(self.data_transformer.vocab),
                      dict_size = len(self.data_transformer.word_vocab),
                      pretrain_dict_embedding = words_embedding,
                      device           = self.device)
        elif (arch == 'bert_lstm'):
            self.model = BERT_LSTM(out_numbers      = out_numbers,
                      model_config     = self.config['models'][arch],
                      device           = self.device)
        elif (arch == 'roberta_lstm'):
            self.model = RoBERTa_LSTM(out_numbers      = out_numbers,
                      model_config     = self.config['models'][arch],
                      device           = self.device)

        self.model = self.model.cuda()
        checkpoint = torch.load(os.path.sep.join([self.config['checkpoint_dir'],self.config['best_model_name'].format(arch=arch)]))
        self.model.load_state_dict(checkpoint['state_dict'])
        predicter = Predicter(model           = self.model,
                          model_name      = arch, 
                          logger          = self.logger,
                          n_gpu           = self.config['n_gpus'],
                          test_data       = test_loader,
                          checkpoint_path = self.checkpoint_path,
                          label_to_id     = self.config['multi-task'][0]["label_to_id"],
                          distributed      = self.is_distributed)


        predictions = predicter.predict()
        for task,pred in zip(self.config['multi-task'],predictions):
            result_path = task['result_path']
            raw_test_path = task['raw_test_path']
            taskname = task['name']
            test_write(data = pred,filename = result_path,raw_text_path=task['raw_test_path'], name=taskname)
            print("current task is %s and output file is at %s"%(taskname,result_path))
        if len(config['n_gpus']) > 0:
            torch.cuda.empty_cache()

   


#encoding:utf-8
import os
import logging
from logging import Logger
from .utils import ensure_dir
from logging.handlers import TimedRotatingFileHandler

'''
日志模块
1. 同时将日志打印到屏幕跟文件中
2. 默认值保留近30天日志文件
'''

def init_logger(log_name,log_dir):
    ensure_dir(log_dir)
    if log_name not in Logger.manager.loggerDict:
        logger  = logging.getLogger(log_name)
        logger.setLevel(logging.DEBUG)
        handler = TimedRotatingFileHandler(filename=os.path.join(log_dir,"%s.log"%log_name),when='D',backupCount = 30)
        datefmt = '%Y-%m-%d %H:%M:%S'
        format_str = '[%(asctime)s]: %(name)s %(filename)s[line:%(lineno)s] %(levelname)s  %(message)s'
        formatter = logging.Formatter(format_str,datefmt)
        handler.setFormatter(formatter)
        handler.setLevel(logging.INFO)
        logger.addHandler(handler)
        console= logging.StreamHandler()
        console.setLevel(logging.INFO)
        console.setFormatter(formatter)
        logger.addHandler(console)

        handler = TimedRotatingFileHandler(filename=os.path.join(log_dir,"ERROR.log"),when='D',backupCount= 30)
        datefmt = '%Y-%m-%d %H:%M:%S'
        format_str = '[%(asctime)s]: %(name)s %(filename)s[line:%(lineno)s] %(levelname)s  %(message)s'
        formatter = logging.Formatter(format_str,datefmt)
        handler.setFormatter(formatter)
        handler.setLevel(logging.ERROR)
        logger.addHandler(handler)
    logger = logging.getLogger(log_name)
    return logger

if __name__ == "__main__":
    logger = init_logger('test','D:\data')
    logger.info("test")
    logger.error("test2")


#encoding:utf-8
import os
import sys
import time
import numpy as np
import torch
from itertools import chain
from torch.autograd import Variable
from ..callback.progressbar import ProgressBar
from ..utils.utils import AverageMeter
from .train_utils import restore_checkpoint,model_device
from .metrics import Entity_Score
from .train_utils import batchify_with_label
from ..config.new_config import configs as config


class Trainer(object):
    def __init__(self,model,
                 model_name,
                 train_data,
                 val_data,
                 optimizer,
                 epochs,
                 logger,
                 evaluate,
                 num_tasks = 1,
                 device = None,
                 avg_batch_loss   = False,
                 distributed = True,
                 label_to_id      = None,
                 n_gpu            = None,
                 lr_scheduler     = None,
                 resume           = None,
                 model_checkpoint = None,
                 writer           = None,
                 verbose = 1):
        self.model            = model     
        self.model_name       = model_name      
        self.train_data       = train_data         
        self.val_data         = val_data          
        self.epochs           = epochs          
        self.optimizer        = optimizer   
        self.logger           = logger        
        self.verbose          = verbose            
        self.writer           = writer         
        self.resume           = resume           
        self.model_checkpoint = model_checkpoint 
        self.lr_scheduler     = lr_scheduler     
        self.evaluate         = evaluate         
        self.n_gpu            = n_gpu             
        self.avg_batch_loss   = avg_batch_loss     
        self.num_tasks        = num_tasks
        self.task={0:10,1:3}
        self.id_to_label      = {value:key for key, value in label_to_id.items()}
        self._reset()
        self.device = device

    def _reset(self):

        self.train_entity_score = Entity_Score(id_to_label=self.id_to_label)
        self.val_entity_score   = Entity_Score(id_to_label=self.id_to_label)

        self.batch_num         = len(self.train_data)
        self.progressbar       = ProgressBar(n_batch = self.batch_num,eval_name='acc',loss_name='loss')
        self.model,self.device = model_device(n_gpu=self.n_gpu,model = self.model,logger = self.logger)
        self.start_epoch = 1

        if self.resume:
            arch = self.model_checkpoint.arch
            resume_path = os.path.join(self.model_checkpoint.checkpoint_dir.format(arch = arch),
                                       self.model_checkpoint.best_model_name.format(arch = arch))
            self.logger.info("\nLoading checkpoint: {} ...".format(resume_path))
            resume_list = restore_checkpoint(resume_path = resume_path,model = self.model,optimizer = self.optimizer)
            self.model     = resume_list[0]
            self.optimizer = resume_list[1]
            best           = resume_list[2]
            self.start_epoch = resume_list[3]
        

            if self.model_checkpoint:
                self.model_checkpoint.best = best
            self.logger.info("\nCheckpoint '{}' (epoch {}) loaded".format(resume_path, self.start_epoch))

    def summary(self):
        model_parameters = filter(lambda p: p.requires_grad, self.model.module.parameters())

        params = sum([np.prod(p.size()) for p in model_parameters])

        self.logger.info('trainable parameters: {:4}M'.format(params / 1000 / 1000))



    def _save_info(self,epoch,val_loss):
        state = {
            'epoch': epoch,
            'arch': self.model_checkpoint.arch,
            'state_dict': self.model.module.state_dict(),
            'optimizer': self.optimizer.state_dict(),
            'val_loss': round(val_loss,4)
        }
        return state
        
    def sequence_mask(self, sequence_length, max_len, device=None):   # sequence_length :(batch_size, )
        batch_size = sequence_length.size(0)     
        seq_range = torch.arange(0, max_len).long()
        seq_range_expand = seq_range.unsqueeze(0).expand(batch_size, max_len)
        seq_range_expand = Variable(seq_range_expand)
        if sequence_length.is_cuda:
            seq_range_expand = seq_range_expand.cuda()
        seq_length_expand = (sequence_length.unsqueeze(1).expand_as(seq_range_expand))
        return seq_range_expand < seq_length_expand 

    def _valid_epoch(self):
        self.model.module.eval()
        val_losses_ner = AverageMeter()
        val_losses_cws = AverageMeter()#!
        val_acc_ner    = AverageMeter()
        val_acc_cws = AverageMeter()#!
        val_f1     = AverageMeter()
        
        

        for idx,(inputs,gaz,target,length,source) in enumerate(self.val_data):
            check = sum(source.numpy().tolist())
            if (check%self.num_tasks!=0):
                continue

            inputs = (torch.stack(inputs,dim=0)).t().cuda()

            target = (torch.stack(target,dim=0)).t().cuda()

            batch_size = inputs.size(0)

            masks = self.sequence_mask(length, config['max_length'],self.device) if ('bert' in self.model_name) else None
            train_task = self.task[check/batch_size]

            outputs = self.model(inputs, length ,out_number = train_task, masks=masks, gaz=gaz)
            mask,target = batchify_with_label(inputs = inputs,target = target,outputs = outputs)
            loss = self.model.module.crf.neg_log_likelihood_loss(outputs, mask,target,out_number = train_task) 
            if self.avg_batch_loss:
                loss /=  batch_size
            _,predicts = self.model.module.crf(outputs, mask,out_number = train_task)
                    
            acc,f1 = self.evaluate(predicts,target=target)
            if (check/batch_size==0):
                val_losses_ner.update(loss.item(),batch_size)
                val_acc_ner.update(acc.item(),batch_size)

            else:
                val_losses_cws.update(loss.item(),batch_size)
                val_acc_cws.update(acc.item(),batch_size)
            val_f1.update(f1.item(),batch_size)
            if self.device != 'cpu':
                predicts = predicts.cpu().numpy()
                target = target.cpu().numpy()
            if (check/batch_size==0):
                self.val_entity_score.update(pred_paths=predicts, label_paths=target)


        return {
            'val_loss': val_losses_ner.avg,
            'val_loss2': val_losses_cws.avg,
            'val_acc': val_acc_ner.avg,
            'val_acc2': val_acc_cws.avg,
            'val_f1': val_f1.avg
        }



    def _train_epoch(self):
        self.model.module.train()
        train_loss_ner = AverageMeter()
        train_acc_ner  = AverageMeter()
        train_f1   = AverageMeter()
        train_loss_cws = AverageMeter()
        train_acc_cws  = AverageMeter()

                
        batch_loss_ner = 0
        acc_ner=0
        acc_cws=0
        batch_loss_cws = 0 


        for idx,(inputs,gaz,target,length,source) in enumerate(self.train_data):
            check = sum(source.numpy().tolist())
            if (check%self.num_tasks!=0):
                continue

            inputs = (torch.stack(inputs,dim=0)).t().cuda()
            target = (torch.stack(target,dim=0)).t().cuda()
            batch_size = inputs.size(0)
            train_task = self.task[check/batch_size]

           # gaz = gaz.cuda()        
 
            masks = self.sequence_mask(length, config['max_length'],self.device) if ('bert' in self.model_name) else None

            outputs = self.model(inputs, length, out_number = train_task, masks = masks, gaz=gaz)

            mask, target = batchify_with_label(inputs=inputs, target=target, outputs=outputs)
            loss    = self.model.module.crf.neg_log_likelihood_loss(outputs,mask,target, out_number = train_task)
            if self.avg_batch_loss:
                loss  /= batch_size
            _,predicts = self.model.module.crf(outputs,mask,out_number = train_task)
            acc,f1 = self.evaluate(predicts,target)

            if (not 'lattice' in self.model_name):
                loss.backward()
                self.optimizer.step()
                self.optimizer.zero_grad()
            else:
            	assert(1+1==3)
            if (check/batch_size==0):
                train_loss_ner.update(loss.item(),batch_size)
                train_acc_ner.update(acc.item(),batch_size)
                train_f1.update(f1.item(),batch_size)
            else:
                train_loss_cws.update(loss.item(),batch_size)
                train_acc_cws.update(acc.item(),batch_size)


            if self.device != 'cpu':
                predicts = predicts.cpu().numpy()
                target   = target.cpu().numpy()
                #self.train_entity_score.update(pred_paths=predicts,label_paths=target)
        
            if (self.verbose >= 1 and (not 'lattice' in self.model_name)):
                self.progressbar.step(batch_idx=idx,
                                      loss     = loss.item(),
                                      acc      = acc.item(),
                                      f1       = f1.item(),
                                      use_time = 0)

        train_log = {
            'loss': train_loss_ner.avg,
            'acc': train_acc_ner.avg,
            'loss2': train_loss_cws.avg,
            'acc2': train_acc_cws.avg,
            'f1': train_f1.avg,
        }
        return train_log

    def train(self):
        for epoch in range(self.start_epoch,self.start_epoch+self.epochs):
            print("----------------- training start -----------------------")
            print("Epoch {i}/{epochs}......".format(i=epoch, epochs=self.start_epoch+self.epochs -1))

            train_log = self._train_epoch()
            val_log = self._valid_epoch()

            logs = dict(train_log,**val_log)

            self.val_entity_score.result()
        
            
            self.logger.info('\nEpoch: %d - loss: %.4f acc: %.4f - f1: %.4f val_loss: %.4f - val_acc: %.4f - val_f1: %.4f - val_acc2: %.4f'%(
                            epoch,logs['loss'],logs['acc'],logs['f1'],logs['val_loss'],logs['val_acc'],logs['val_f1'],logs['val_acc2'])
                             )

            if self.lr_scheduler:
                self.lr_scheduler.step(logs['loss'],epoch)


            if self.model_checkpoint:
                state = self._save_info(epoch,val_loss = logs['val_loss'])
                self.model_checkpoint.step(current=logs[self.model_checkpoint.monitor],state = state)



# -*- coding: utf-8 -*-
# @Author: Jie
# @Date:   2017-06-15 14:23:06
# @Last Modified by:   Jie Yang,     Contact: jieynlp@gmail.com
# @Last Modified time: 2018-05-12 22:09:37
import sys
import numpy as np
from utils.alphabet import Alphabet
NULLKEY = "-null-"
def normalize_word(word):
    new_word = ""
    for char in word:
        if char.isdigit():
            new_word += '0'
        else:
            new_word += char
    return new_word


def read_instance(input_file, word_alphabet, char_alphabet, label_alphabet, number_normalized,max_sent_length, char_padding_size=-1, char_padding_symbol = '</pad>'):
    in_lines = open(input_file,'r',encoding='utf-8').readlines()
    instence_texts = []
    instence_Ids = []
    words = []
    chars = []
    labels = []
    word_Ids = []
    char_Ids = []
    label_Ids = []
    for line in in_lines:
        if len(line) > 2:
            pairs = line.strip().split()
            word = pairs[0]
            if number_normalized:
                word = normalize_word(word)
            label = pairs[-1]
            words.append(word)
            labels.append(label)
            word_Ids.append(word_alphabet.get_index(word))
            label_Ids.append(label_alphabet.get_index(label))
            char_list = []
            char_Id = []
            for char in word:
                char_list.append(char)
            if char_padding_size > 0:
                char_number = len(char_list)
                if char_number < char_padding_size:
                    char_list = char_list + [char_padding_symbol]*(char_padding_size-char_number)
                assert(len(char_list) == char_padding_size)
            else:
                ### not padding
                pass
            for char in char_list:
                char_Id.append(char_alphabet.get_index(char))
            chars.append(char_list)
            char_Ids.append(char_Id)
        else:
            if (max_sent_length < 0) or (len(words) < max_sent_length):
                instence_texts.append([words, chars, labels])
                instence_Ids.append([word_Ids, char_Ids,label_Ids])
            words = []
            chars = []
            labels = []
            word_Ids = []
            char_Ids = []
            label_Ids = []
    return instence_texts, instence_Ids


def read_seg_instance(input_file, word_alphabet, biword_alphabet, char_alphabet, label_alphabet, number_normalized, max_sent_length, char_padding_size=-1, char_padding_symbol = '</pad>'):
    in_lines = open(input_file,'r',encoding='utf-8').readlines()
    instence_texts = []
    instence_Ids = []
    words = []
    biwords = []
    chars = []
    labels = []
    word_Ids = []
    biword_Ids = []
    char_Ids = []
    label_Ids = []
    for idx in xrange(len(in_lines)):
        line = in_lines[idx]
        if len(line) > 2:
            pairs = line.strip().split()
            word = pairs[0]
            if number_normalized:
                word = normalize_word(word)
            label = pairs[-1]
            words.append(word)
            if idx < len(in_lines) -1 and len(in_lines[idx+1]) > 2:
                biword = word + in_lines[idx+1].strip().split()[0]
            else:
                biword = word + NULLKEY
            biwords.append(biword)
            labels.append(label)
            word_Ids.append(word_alphabet.get_index(word))
            biword_Ids.append(biword_alphabet.get_index(biword))
            label_Ids.append(label_alphabet.get_index(label))
            char_list = []
            char_Id = []
            for char in word:
                char_list.append(char)
            if char_padding_size > 0:
                char_number = len(char_list)
                if char_number < char_padding_size:
                    char_list = char_list + [char_padding_symbol]*(char_padding_size-char_number)
                assert(len(char_list) == char_padding_size)
            else:
                ### not padding
                pass
            for char in char_list:
                char_Id.append(char_alphabet.get_index(char))
            chars.append(char_list)
            char_Ids.append(char_Id)
        else:
            if (max_sent_length < 0) or (len(words) < max_sent_length):
                instence_texts.append([words, biwords, chars, labels])
                instence_Ids.append([word_Ids, biword_Ids, char_Ids,label_Ids])
            words = []
            biwords = []
            chars = []
            labels = []
            word_Ids = []
            biword_Ids = []
            char_Ids = []
            label_Ids = []
    return instence_texts, instence_Ids


def read_instance_with_gaz(input_file, gaz, word_alphabet, biword_alphabet, char_alphabet, gaz_alphabet, label_alphabet, segment_label_alphabet, number_normalized, max_sent_length, mode='NER', char_padding_size=-1, char_padding_symbol = '</pad>'):
    in_lines = open(input_file,'r',encoding='utf-8').readlines()
    instence_texts = []
    instence_Ids = []
    words = []
    biwords = []
    chars = []
    labels = []
    word_Ids = []
    biword_Ids = []
    char_Ids = []
    label_Ids = []
    for idx in range(len(in_lines)):
        line = in_lines[idx]
        if len(line) > 2:
            pairs = line.strip().split()
            word = pairs[0]
            if number_normalized:
                word = normalize_word(word)
            label = pairs[-1]
         ###   print("CHECK:",word, label)
            if idx < len(in_lines) -1 and len(in_lines[idx+1]) > 2:
                biword = word + in_lines[idx+1].strip().split()[0]
            else:
                biword = word + NULLKEY
            biwords.append(biword)
            words.append(word)
            labels.append(label)
            word_Ids.append(word_alphabet.get_index(word))
            biword_Ids.append(biword_alphabet.get_index(biword))
            if mode == 'NER':
                label_Ids.append(label_alphabet.get_index(label))
            else:
                label_Ids.append(segment_label_alphabet.get_index(label))
                    

            char_list = []
            char_Id = []
            for char in word:
                char_list.append(char)
            if char_padding_size > 0:
                char_number = len(char_list)
                if char_number < char_padding_size:
                    char_list = char_list + [char_padding_symbol]*(char_padding_size-char_number)
                assert(len(char_list) == char_padding_size)
            else:
                ### not padding
                pass
            for char in char_list:
                char_Id.append(char_alphabet.get_index(char))
            chars.append(char_list)
            char_Ids.append(char_Id)

        else:
            if ((max_sent_length < 0) or (len(words) < max_sent_length)) and (len(words)>0):
                gazs = []
                gaz_Ids = []
                w_length = len(words)
                # print sentence 
                # for w in words:
                #     print w," ",
                # print
                for idx in range(w_length):
                    matched_list = gaz.enumerateMatchList(words[idx:])
                    matched_length = [len(a) for a in matched_list]
                    # print idx,"----------"
                    # print "forward...feed:","".join(words[idx:])
                    # for a in matched_list:
                    #     print a,len(a)," ",
                    # print

                    # print matched_length

                    gazs.append(matched_list)
                    matched_Id  = [gaz_alphabet.get_index(entity) for entity in matched_list]
                    if matched_Id:
                        gaz_Ids.append([matched_Id, matched_length])
                    else:
                        gaz_Ids.append([])
                    
                instence_texts.append([words, biwords, chars, gazs, labels])
                instence_Ids.append([word_Ids, biword_Ids, char_Ids, gaz_Ids, label_Ids])
            words = []
            biwords = []
            chars = []
            labels = []
            word_Ids = []
            biword_Ids = []
            char_Ids = []
            label_Ids = []
            gazs = []
            gaz_Ids = []
    return instence_texts, instence_Ids


def read_instance_with_gaz_in_sentence(input_file, gaz, word_alphabet, biword_alphabet, char_alphabet, gaz_alphabet, label_alphabet, number_normalized, max_sent_length, char_padding_size=-1, char_padding_symbol = '</pad>'):
    in_lines = open(input_file,'r',encoding='utf-8').readlines()
    instence_texts = []
    instence_Ids = []
    for idx in range(len(in_lines)):
        pair = in_lines[idx].strip().split()
        orig_words = list(pair[0])
        
        if (max_sent_length > 0) and (len(orig_words) > max_sent_length):
            continue
        biwords = []
        biword_Ids = []
        if number_normalized:
            words = []
            for word in orig_words:
                word = normalize_word(word)
                words.append(word)
        else:
            words = orig_words
        word_num = len(words)
        for idy in range(word_num):
            if idy < word_num - 1:
                biword = words[idy]+words[idy+1]
            else:
                biword = words[idy]+NULLKEY
            biwords.append(biword)
            biword_Ids.append(biword_alphabet.get_index(biword))
        word_Ids = [word_alphabet.get_index(word) for word in words]
        label = pair[-1]
        label_Id =  label_alphabet.get_index(label)
        gazs = []
        gaz_Ids = []
        word_num = len(words)
        chars = [[word] for word in words]
        char_Ids = [[char_alphabet.get_index(word)] for word in words]
        ## print sentence 
        # for w in words:
        #     print w," ",
        # print
        for idx in range(word_num):
            matched_list = gaz.enumerateMatchList(words[idx:])
            matched_length = [len(a) for a in matched_list]
            # print idx,"----------"
            # print "forward...feed:","".join(words[idx:])
            # for a in matched_list:
            #     print a,len(a)," ",
            # print
            # print matched_length
            gazs.append(matched_list)
            matched_Id  = [gaz_alphabet.get_index(entity) for entity in matched_list]
            if matched_Id:
                gaz_Ids.append([matched_Id, matched_length])
            else:
                gaz_Ids.append([])
        instence_texts.append([words, biwords, chars, gazs, label])
        instence_Ids.append([word_Ids, biword_Ids, char_Ids, gaz_Ids, label_Id])
    return instence_texts, instence_Ids


def build_pretrain_embedding(embedding_path, word_alphabet, embedd_dim=100, norm=True):    
    embedd_dict = dict()
    if embedding_path != None:
        embedd_dict, embedd_dim = load_pretrain_emb(embedding_path)
    scale = np.sqrt(3.0 / embedd_dim)
    pretrain_emb = np.empty([word_alphabet.size(), embedd_dim])
    perfect_match = 0
    case_match = 0
    not_match = 0
    for word, index in word_alphabet.iteritems():
        if word in embedd_dict:
            if norm:
                pretrain_emb[index,:] = norm2one(embedd_dict[word])
            else:
                pretrain_emb[index,:] = embedd_dict[word]
            perfect_match += 1
        elif word.lower() in embedd_dict:
            if norm:
                pretrain_emb[index,:] = norm2one(embedd_dict[word.lower()])
            else:
                pretrain_emb[index,:] = embedd_dict[word.lower()]
            case_match += 1
        else:
            pretrain_emb[index,:] = np.random.uniform(-scale, scale, [1, embedd_dim])
            not_match += 1
    pretrained_size = len(embedd_dict)
    print("Embedding:\n     pretrain word:%s, prefect match:%s, case_match:%s, oov:%s, oov%%:%s"%(pretrained_size, perfect_match, case_match, not_match, (not_match+0.)/word_alphabet.size()))
    return pretrain_emb, embedd_dim


       
def norm2one(vec):
    root_sum_square = np.sqrt(np.sum(np.square(vec)))
    return vec/root_sum_square

def load_pretrain_emb(embedding_path):
    embedd_dim = -1
    embedd_dict = dict()
    with open(embedding_path, 'r',encoding='utf-8') as file:
        for line in file:
            line = line.strip()
            if len(line) == 0:
                continue
            tokens = line.split()
            if embedd_dim < 0:
                embedd_dim = len(tokens) - 1
            else:
            #    print(embedd_dim, len(tokens))
                assert (embedd_dim + 1 == len(tokens))
            embedd = np.empty([1, embedd_dim])
            embedd[:] = tokens[1:]
            embedd_dict[tokens[0]] = embedd
    return embedd_dict, embedd_dim

if __name__ == '__main__':
    a = np.arange(9.0)
    print (a)
    print (norm2one(a))



#encoding:utf-8
import numpy as np
import warnings
from torch.optim.optimizer import Optimizer

class StepLr(object):
    def __init__(self,optimizer,lr):
        super(StepLr,self).__init__()
        self.optimizer = optimizer
        self.lr = lr
        self.use = 'on_epoch_end'

    def step(self,epoch):
        lr = self.lr
        if epoch > 12:
            lr = lr / 1000
        elif epoch > 8:
            lr = lr / 100
        elif epoch > 4:
            lr = lr / 10
        for param_group in self.optimizer.param_groups:
            param_group['lr'] = lr

class CustomDecay(object):
    def __init__(self,optimtizer):
        super(CustomDecay,self).__init__()
        self.optimizer = optimtizer

    def step(self,epoch):
        if epoch % 2 != 0:
            for param_group in self.optimizer.param_groups:
                param_group["lr"] = param_group['lr'] * 0.1

class CyclicLR(object):
    def __init__(self, optimizer, base_lr=1e-3, max_lr=6e-3,
                 step_size=2000, mode='triangular', gamma=1.,
                 scale_fn=None, scale_mode='cycle', last_batch_iteration=-1):

        if not isinstance(optimizer, Optimizer):
            raise TypeError('{} is not an Optimizer'.format(
                type(optimizer).__name__))
        self.optimizer = optimizer

        if isinstance(base_lr, list) or isinstance(base_lr, tuple):
            if len(base_lr) != len(optimizer.param_groups):
                raise ValueError("expected {} base_lr, got {}".format(
                    len(optimizer.param_groups), len(base_lr)))
            self.base_lrs = list(base_lr)
        else:
            self.base_lrs = [base_lr] * len(optimizer.param_groups)

        if isinstance(max_lr, list) or isinstance(max_lr, tuple):
            if len(max_lr) != len(optimizer.param_groups):
                raise ValueError("expected {} max_lr, got {}".format(
                    len(optimizer.param_groups), len(max_lr)))
            self.max_lrs = list(max_lr)
        else:
            self.max_lrs = [max_lr] * len(optimizer.param_groups)

        self.step_size = step_size

        if mode not in ['triangular', 'triangular2', 'exp_range'] \
                and scale_fn is None:
            raise ValueError('mode is invalid and scale_fn is None')

        self.mode = mode
        self.gamma = gamma

        if scale_fn is None:
            if self.mode == 'triangular':
                self.scale_fn = self._triangular_scale_fn
                self.scale_mode = 'cycle'
            elif self.mode == 'triangular2':
                self.scale_fn = self._triangular2_scale_fn
                self.scale_mode = 'cycle'
            elif self.mode == 'exp_range':
                self.scale_fn = self._exp_range_scale_fn
                self.scale_mode = 'iterations'
        else:
            self.scale_fn = scale_fn
            self.scale_mode = scale_mode

        self.batch_step(last_batch_iteration + 1)
        self.last_batch_iteration = last_batch_iteration

    def _triangular_scale_fn(self, x):
        return 1.

    def _triangular2_scale_fn(self, x):
        return 1 / (2. ** (x - 1))

    def _exp_range_scale_fn(self, x):
        return self.gamma**(x)

    def get_lr(self):
        step_size = float(self.step_size)
        cycle = np.floor(1 + self.last_batch_iteration / (2 * step_size))
        x = np.abs(self.last_batch_iteration / step_size - 2 * cycle + 1)

        lrs = []
        param_lrs = zip(self.optimizer.param_groups, self.base_lrs, self.max_lrs)
        for param_group, base_lr, max_lr in param_lrs:
            base_height = (max_lr - base_lr) * np.maximum(0, (1 - x))
            if self.scale_mode == 'cycle':
                lr = base_lr + base_height * self.scale_fn(cycle)
            else:
                lr = base_lr + base_height * self.scale_fn(self.last_batch_iteration)
            lrs.append(lr)
        return lrs

    def batch_step(self, batch_iteration=None):
        if batch_iteration is None:
            batch_iteration = self.last_batch_iteration + 1
        self.last_batch_iteration = batch_iteration
        for param_group, lr in zip(self.optimizer.param_groups, self.get_lr()):
            param_group['lr'] = lr

class ReduceLROnPlateau(object):
    """Reduce learning rate when a metric has stopped improving.
    Models often benefit from reducing the learning rate by a factor
    of 2-10 once learning stagnates. This scheduler reads a metrics
    quantity and if no improvement is seen for a 'patience' number
    of epochs, the learning rate is reduced.

    Args:
        factor: factor by which the learning rate will
            be reduced. new_lr = lr * factor
        patience: number of epochs with no improvement
            after which learning rate will be reduced.
        verbose: int. 0: quiet, 1: update messages.
        mode: one of {min, max}. In `min` mode,
            lr will be reduced when the quantity
            monitored has stopped decreasing; in `max`
            mode it will be reduced when the quantity
            monitored has stopped increasing.
        epsilon: threshold for measuring the new optimum,
            to only focus on significant changes.
        cooldown: number of epochs to wait before resuming
            normal operation after lr has been reduced.
        min_lr: lower bound on the learning rate.


    Example:
        >>> optimizer = torch.optim.SGD(model.parameters(), lr=0.1, momentum=0.9)
        >>> scheduler = ReduceLROnPlateau(optimizer, 'min')
        >>> for epoch in range(10):
        >>>     train(...)
        >>>     val_acc, val_loss = validate(...)
        >>>     scheduler.step(val_loss, epoch)
    """

    def __init__(self, optimizer, mode='min', factor=0.1, patience=10,
                 verbose=0, epsilon=1e-4, cooldown=0, min_lr=0):
        super(ReduceLROnPlateau, self).__init__()

        if factor >= 1.0:
            raise ValueError('ReduceLROnPlateau '
                             'does not support a factor >= 1.0.')
        self.factor = factor
        self.min_lr = min_lr
        self.epsilon = epsilon
        self.patience = patience
        self.verbose = verbose
        self.cooldown = cooldown
        self.cooldown_counter = 0  # Cooldown counter.
        self.monitor_op = None
        self.wait = 0
        self.best = 0
        self.mode = mode
        assert isinstance(optimizer, Optimizer)
        self.optimizer = optimizer
        self.use = 'on_epoch_end'
        self._reset()

    def _reset(self):
        """Resets wait counter and cooldown counter.
        """
        if self.mode not in ['min', 'max']:
            raise RuntimeError('Learning Rate Plateau Reducing mode %s is unknown!')
        if self.mode == 'min':
            self.monitor_op = lambda a, b: np.less(a, b - self.epsilon)
            self.best = np.Inf
        else:
            self.monitor_op = lambda a, b: np.greater(a, b + self.epsilon)
            self.best = -np.Inf
        self.cooldown_counter = 0
        self.wait = 0
        self.lr_epsilon = self.min_lr * 1e-4

    def reset(self):
        self._reset()

    def step(self, metrics, epoch):
        current = metrics
        if current is None:
            warnings.warn('Learning Rate Plateau Reducing requires metrics available!', RuntimeWarning)
        else:
            if self.in_cooldown():
                self.cooldown_counter -= 1
                self.wait = 0

            if self.monitor_op(current, self.best):
                self.best = current
                self.wait = 0
            elif not self.in_cooldown():
                if self.wait >= self.patience:
                    for param_group in self.optimizer.param_groups:
                        old_lr = float(param_group['lr'])
                        if old_lr > self.min_lr + self.lr_epsilon:
                            new_lr = old_lr * self.factor
                            new_lr = max(new_lr, self.min_lr)
                            param_group['lr'] = new_lr
                            if self.verbose > 0:
                                print('\nEpoch %05d: reducing learning rate to %s.' % (epoch, new_lr))
                            self.cooldown_counter = self.cooldown
                            self.wait = 0
                self.wait += 1

    def in_cooldown(self):
        return self.cooldown_counter > 0



#encoding:utf-8
import json
from torchtext import data
from ..utils.gazetteer import Gazetteer
from ..config.new_config import configs as config
from ..config.new_config import BASE_DIR
import numpy
from transformers import BertTokenizer
from torch.utils.data.distributed import DistributedSampler
from torch.utils.data import DataLoader
import os
from tqdm import tqdm
import re
import torch



class MyTokenizer(object):
    def __init__(self, pretrained_model_path=None, mask_entity=False):
        self.pretrained_model_path = pretrained_model_path or 'bert-base-chinese'
        self.bert_tokenizer = BertTokenizer.from_pretrained(self.pretrained_model_path)
        self.mask_entity = mask_entity
    def group_tokenize(self, item):
        etts = []
        sentence = item['text']
        info = item['info']
        if (len(info)<2):
            return [],[]
        results = []
        for head_idx in range(len(info)-1):
            tail_idx = head_idx + 1
            head_dict = {}
            head_dict['name'] = info[head_idx]['word']

            head_dict['pos'] = [info[head_idx]['start']]
            head_dict['pos'].append(info[head_idx]['end']+1)

            while tail_idx<len(info):
                tail_dict = {}
                tail_dict['name'] = info[tail_idx]['word']
                tail_dict['pos'] = [info[tail_idx]['start']]
                tail_dict['pos'].append(info[tail_idx]['end']+1)
                etts.append({'head':head_dict,'tail':tail_dict})
                etts.append({'tail':head_dict,'head':tail_dict})
                query1 = {'h':head_dict,'t':tail_dict,'text':sentence}
                query2 = {'h':tail_dict,'t':head_dict,'text':sentence}
                token,e1,e2 = self.tokenize(query1)
                results.append({'token':token,'e1':e1,'e2':e2})
                token,e1,e2 = self.tokenize(query2)
                results.append({'token':token,'e1':e1,'e2':e2})
                tail_idx+=1

        return results, etts

    def tokenize(self, item):
        sentence = item['text']
        pos_head = item['h']['pos']
        pos_tail = item['t']['pos']

        if pos_head[0] > pos_tail[0]:
            pos_min = pos_tail
            pos_max = pos_head
            rev = True
        else:
            pos_min = pos_head
            pos_max = pos_tail
            rev = False

        sent0 = self.bert_tokenizer.tokenize(sentence[:pos_min[0]])
        ent0 = self.bert_tokenizer.tokenize(sentence[pos_min[0]:pos_min[1]])
        sent1 = self.bert_tokenizer.tokenize(sentence[pos_min[1]:pos_max[0]])
        ent1 = self.bert_tokenizer.tokenize(sentence[pos_max[0]:pos_max[1]])
        sent2 = self.bert_tokenizer.tokenize(sentence[pos_max[1]:])

        if rev:
            if self.mask_entity:
                ent0 = ['[unused6]']
                ent1 = ['[unused5]']
            pos_tail = [len(sent0), len(sent0) + len(ent0)]
            pos_head = [
                len(sent0) + len(ent0) + len(sent1),
                len(sent0) + len(ent0) + len(sent1) + len(ent1)
            ]
        else:
            if self.mask_entity:
                ent0 = ['[unused5]']
                ent1 = ['[unused6]']
            pos_head = [len(sent0), len(sent0) + len(ent0)]
            pos_tail = [
                len(sent0) + len(ent0) + len(sent1),
                len(sent0) + len(ent0) + len(sent1) + len(ent1)
            ]
        tokens = sent0 + ent0 + sent1 + ent1 + sent2

        re_tokens = ['[CLS]']
        cur_pos = 0
        pos1 = [0, 0]
        pos2 = [0, 0]
        for token in tokens:
            token = token.lower()
            if cur_pos == pos_head[0]:
                pos1[0] = len(re_tokens)
                re_tokens.append('[unused1]')
            if cur_pos == pos_tail[0]:
                pos2[0] = len(re_tokens)
                re_tokens.append('[unused2]')
            re_tokens.append(token)
            if cur_pos == pos_head[1] - 1:
                re_tokens.append('[unused3]')
                pos1[1] = len(re_tokens)
            if cur_pos == pos_tail[1] - 1:
                re_tokens.append('[unused4]')
                pos2[1] = len(re_tokens)
            cur_pos += 1
        re_tokens.append('[SEP]')
        return re_tokens[1:-1], pos1, pos2



class REDataset(data.Dataset):
    def __init__(self, data_file_path, tagset_path, pretrained_model_path=None, max_len=128, ner2re=False):
        if ner2re:
            self.data_file_path = data_file_path
            self.tagset_path = tagset_path
            self.pretrained_model_path = pretrained_model_path or 'bert-base-chinese'
            self.tokenizer = MyTokenizer(pretrained_model_path=self.pretrained_model_path)
            self.max_len = max_len
            self.tokens_list, self.e1_mask_list,self.e2_mask_list, self.tags, self.etts = trans(data_file_path,tokenizer=self.tokenizer,max_len=self.max_len)
            self.tag2idx = get_tag2idx(self.tagset_path)
        else:    
            self.data_file_path = data_file_path
            self.tagset_path = tagset_path
            self.pretrained_model_path = pretrained_model_path or 'bert-base-chinese'
            self.tokenizer = MyTokenizer(pretrained_model_path=self.pretrained_model_path)
            self.max_len = max_len
            self.tokens_list, self.e1_mask_list, self.e2_mask_list, self.tags = read_data(data_file_path, tokenizer=self.tokenizer, max_len=self.max_len)
            self.tag2idx = get_tag2idx(self.tagset_path)


    def __len__(self):
        return len(self.tags)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        sample_tokens = self.tokens_list[idx]
        sample_e1_mask = self.e1_mask_list[idx]
        sample_e2_mask = self.e2_mask_list[idx]
        sample_tag = self.tags[idx]
        encoded = self.tokenizer.bert_tokenizer.encode_plus(sample_tokens, max_length=self.max_len, pad_to_max_length=True)
        sample_token_ids = encoded['input_ids']
        sample_token_type_ids = encoded['token_type_ids']
        sample_attention_mask = encoded['attention_mask']
        try:
            sample_tag_id = torch.tensor(self.tag2idx[sample_tag])
        except:
            sample_tag_id = sample_tag
        sample = {
            'token_ids': torch.tensor(sample_token_ids),
            'token_type_ids': torch.tensor(sample_token_type_ids),
            'attention_mask': torch.tensor(sample_attention_mask),
            'e1_mask': torch.tensor(sample_e1_mask),
            'e2_mask': torch.tensor(sample_e2_mask),
            'tag_id': sample_tag_id
        }
        return sample

class MultiDataset(data.Dataset):
    def __init__(self,
                num_threads,
                 data_paths,
                 x_var,y_var,
                 skip_header,
                 is_train_mode,
                 batch_size,
                 max_sentence_length = None,
                 default_token = False,
                 gaz = None,
                 vocab = None,
                 rev_vocab = None,

                 ):
        self.data_group = []
        self.data = {'sents':[],'labels':[],'gaz_lists':[],'lengths':[],'sources':[]}
        if (default_token):
            t = BertTokenizer.from_pretrained(BASE_DIR+'/output/checkpoints/roberta')
            tokenizer = lambda x: t.convert_tokens_to_ids(x.split())
        else:
            tokenizer = lambda x: [int(c) for c in x.split()]
        label_tok = lambda x: [int(c) for c in x.split()]
        padding = lambda x: [(0 if idx>=len(x) else x[idx]) for idx in range(max_sentence_length)]
        for idx,data_path in enumerate(data_paths):
            self.data_group.append({})
            sents = []
            labels = []
            gaz_lists = []
            lengths = []
            sources = []
            with open(data_path,'r') as fr:
                for i,line in enumerate(fr):
                    if i == 0 and skip_header:
                        continue
                    df = json.loads(line)
                    sentence = df[x_var]
                    label = df[y_var]

                    sents.append(padding(tokenizer(sentence)))

                    labels.append(padding(label_tok(label)))
                    lengths.append(len(sentence.split()) if (len(sentence.split())<max_sentence_length) else max_sentence_length)
                    sources.append(idx)
            self.data_group[idx]['sents']=sents
            self.data_group[idx]['labels']=labels
            self.data_group[idx]['gaz_lists']=gaz_lists
            self.data_group[idx]['lengths']=lengths
            self.data_group[idx]['sources']=sources
        offset = 0
        batch_sz = batch_size*num_threads
        while True:
            endoffile = True
            for group in self.data_group:
                for k in group.keys():
                    d = group[k]

                    if (offset+batch_sz<len(d)):
                        self.data[k].extend(d[offset:offset+batch_sz])
                        endoffile = False
            offset+=batch_sz
            if endoffile:
                break

        if gaz!=None:
            for i,sent in enumerate(self.data['sents']):
                gaz_pack = []
                sent_len = self.data['lengths'][i]          
                for idx in range(sent_len):
                    match_out = gaz.enumerateMatchList(sent[idx:sent_len])
                    if (len(match_out[0])==0):
                        match_out = []
                    gaz_pack.append(match_out)
                while (len(gaz)<max_sentence_length):
                    gaz.append([])
                self.data['gaz_lists'].append(gaz)
        else:
            self.data['gaz_lists']= [0]*len(self.data['sents'])
    def __len__(self):
        return len(self.data['sents'])
    def __getitem__(self,index):
        return (self.data['sents'][index],self.data['gaz_lists'][index],self.data['labels'][index],self.data['lengths'][index],self.data['sources'][index])      



def convert_pos_to_mask(e_pos, max_len=128):
    e_pos_mask = [0] * max_len
    for i in range(e_pos[0], e_pos[1]):
        e_pos_mask[i] = 1
    return e_pos_mask

def trans(input_file,tokenizer=None,max_len=128):
    tokens_list = []
    e1_mask_list = []
    e2_mask_list = []
    tags = []

    with open(input_file, 'r', encoding='utf-8') as f_in:
        liness = 0
        etts = []
        for line in tqdm(f_in):
            liness+=1
            line = line.strip()
            item = json.loads(line)
            if tokenizer is None:
                tokenizer = MyTokenizer()
            results, _etts = tokenizer.group_tokenize(item)
            etts.extend(_etts)
            for result in results:

                tokens = result['token']
                pos_e1 = result['e1']
                pos_e2 = result['e2']
                if pos_e1[0] < max_len - 1 and pos_e1[1] < max_len and \
                        pos_e2[0] < max_len - 1 and pos_e2[1] < max_len:
                    tokens_list.append(tokens)
                    e1_mask = convert_pos_to_mask(pos_e1, max_len)
                    e2_mask = convert_pos_to_mask(pos_e2, max_len)
                    e1_mask_list.append(e1_mask)
                    e2_mask_list.append(e2_mask)
                    tags.append(item['text'])
    return tokens_list, e1_mask_list, e2_mask_list,tags, etts

def read_data(input_file, tokenizer=None, max_len=128):
    tokens_list = []
    e1_mask_list = []
    e2_mask_list = []
    tags = []
    with open(input_file, 'r', encoding='utf-8') as f_in:
        liness = 0
        for line in tqdm(f_in):
            liness+=1
            #if (liness>180000):
            #    break
            line = line.strip()
            item = json.loads(line)
            if tokenizer is None:
                tokenizer = MyTokenizer()
            tokens, pos_e1, pos_e2 = tokenizer.tokenize(item)
            if pos_e1[0] < max_len - 1 and pos_e1[1] < max_len and \
                    pos_e2[0] < max_len - 1 and pos_e2[1] < max_len:
                tokens_list.append(tokens)
                e1_mask = convert_pos_to_mask(pos_e1, max_len)
                e2_mask = convert_pos_to_mask(pos_e2, max_len)
                e1_mask_list.append(e1_mask)
                e2_mask_list.append(e2_mask)
                tag = item['relation']
                tags.append(tag)
    return tokens_list, e1_mask_list, e2_mask_list, tags


def save_tagset(tagset, output_file):
    with open(output_file, 'w', encoding='utf-8') as f_out:
        f_out.write('\n'.join(tagset))


def get_tag2idx(file):
    with open(file, 'r', encoding='utf-8') as f_in:
        tagset = re.split(r'\s+', f_in.read().strip())
    return dict((tag, idx) for idx, tag in enumerate(tagset))


def get_idx2tag(file):
    with open(file, 'r', encoding='utf-8') as f_in:
        tagset = re.split(r'\s+', f_in.read().strip())
    return dict((idx, tag) for idx, tag in enumerate(tagset))





import sys
import requests
import os

requests.packages.urllib3.disable_warnings()

class Downloader(object):
    def __init__(self,prog=True):
        self.prog = prog
        self.config = {'embedding':'https://cloud.tsinghua.edu.cn/f/258e62ed039e46978717/?dl=1','roberta':'https://cloud.tsinghua.edu.cn/f/f2facb2e4bbc4406bbb5/?dl=1',
        'bilstm_crf':'','bert_lstm':'','roberta_lstm':'','REdataset':'https://cloud.tsinghua.edu.cn/f/2f2bc23d9d4d41c0b287/?dl=1'}
    def download(self,url, file_path):
        
        r = requests.get(self.config[url], stream=True, verify=False)


        total_size = int(r.headers['Content-Length'])
        temp_size = 0
        with open(file_path, "wb") as f:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    temp_size += len(chunk)
                    f.write(chunk)
                    f.flush()
                    if self.prog:
                        done = int(50 * temp_size / total_size)
                        sys.stdout.write("\rdownloading %s : [%s%s] %d%%" % (url, '█' * done, ' ' * (50 - done), 100 * temp_size / total_size))
                        sys.stdout.flush()
        print() 




#encoding:utf-8
import torch
from torch.autograd import Variable
from tqdm import tqdm
from .predict_utils import get_entity, get_word
from ..train.train_utils import restore_checkpoint,model_device
from ..train.trainer import batchify_with_label
from ..config.new_config import configs as config

class Predicter(object):
    def __init__(self,
                 model,
                 model_name,
                 test_data,
                 logger,
                 label_to_id,
                 checkpoint_path,
                 n_gpu = 0,
                 distributed = True):
        self.model           = model
        self.model_name      = model_name
        self.test_data       = test_data
        self.logger          = logger
        self.checkpoint_path = checkpoint_path
        self.n_gpu           = n_gpu
        self.task={0:10,1:3}
        self.id2name = {0:"NER",1:"CWS"}
        self.id_to_label      = {value:key for key, value in label_to_id.items()}

    def sequence_mask(self, sequence_length, max_len): 
        batch_size = sequence_length.size(0)            
        seq_range = torch.arange(0, max_len).long()
        seq_range_expand = seq_range.unsqueeze(0).expand(batch_size, max_len)
        seq_range_expand = Variable(seq_range_expand)
        if sequence_length.is_cuda:
            seq_range_expand = seq_range_expand.cuda()
        seq_length_expand = (sequence_length.unsqueeze(1).expand_as(seq_range_expand))
        return seq_range_expand < seq_length_expand 
        


    def _predict_batch(self,inputs,gaz,length, task_number, task):
        with torch.no_grad():
            masks = self.sequence_mask(length, config['max_length']) if ('bert' in self.model_name) else None
            outputs = self.model(inputs, length,task_number,masks, gaz)
            mask, _ = batchify_with_label(inputs=inputs, outputs=outputs,is_train_mode=False)
            _,predicts = self.model.crf(outputs, mask, out_number = task_number)

            batch_result = []
            for index,(text,path) in enumerate(zip(inputs,predicts)):
                try:
                    path = path.cpu().numpy()
                except:
                    pass
                result = get_entity(path = path,tag_map=self.id_to_label) if (task=='NER') else get_word(path=path)
                batch_result.append(result)
            return batch_result

    def predict(self):
        self.model.eval()
        predictions = []
        taskDict = {}
        for idx,(inputs,gaz,target,length,source) in enumerate(self.test_data):
            check = sum(source.numpy().tolist())
            if (check%len(self.task.keys())!=0):
                continue
            inputs = (torch.stack(inputs,dim=0)).t().cuda()
            length = length.cuda()
            batch_size = inputs.size(0)
            cur_task = self.task[check/batch_size]
            cur_name = self.id2name[check/batch_size]


            y_pred_batch = self._predict_batch(inputs = inputs,length = length, gaz = gaz, task_number = cur_task, task=cur_name)
            if (check/batch_size in taskDict.keys()):
                predictions[taskDict[check/batch_size]].extend(y_pred_batch)
            else:
                d = len(taskDict.keys())
                taskDict[check/batch_size]=d
                predictions.append([])
                predictions[taskDict[check/batch_size]].extend(y_pred_batch)
        return predictions


#encoding:utf-8
import torch
import torch.nn.functional as F

# 多分类
class CrossEntropy(object):
    def __init__(self):
        super(CrossEntropy,self).__init__()
        pass
    def __call__(self, output, target):
        loss = F.cross_entropy(input=output, target=target)
        return loss

# 二分类
# define binary cross entropy loss
# note that the model returns logit to take advantage of the log-sum-exp trick
# for numerical stability in the loss
binary_loss = torch.nn.BCEWithLogitsLoss(reduction='sum')


from .trie import Trie

class Gazetteer:
    def __init__(self, lower):
        self.trie = Trie()
        self.ent2type = {} ## word list to type
        self.ent2id = {"<UNK>":0}   ## word list to id
        self.lower = lower
        self.space = ""

    def enumerateMatchList(self, word_list):
        match_list = self.trie.enumerateMatch(word_list, self.space)
        return match_list

    def insert(self, word_list, num):
        #if self.lower:
        #    word_list = [word.lower() for word in word_list]
        self.trie.insert(word_list, num)
        #string = self.space.join(word_list)
        #if string not in self.ent2type:
        #    self.ent2type[string] = source
        #if string not in self.ent2id:
        #    self.ent2id[string] = len(self.ent2id)

    def searchId(self, word_list):
        if self.lower:
            word_list = [word.lower() for word in word_list]
        string = self.space.join(word_list)
        if string in self.ent2id:
            return self.ent2id[string]
        return self.ent2id["<UNK>"]

    def searchType(self, word_list):
        if self.lower:
            word_list = [word.lower() for word in word_list]
        string = self.space.join(word_list)
        if string in self.ent2type:
            return self.ent2type[string]
        print  ("Error in finding entity type at gazetteer.py, exit program! String:", string)
        exit(0)

    def size(self):
        return len(self.ent2type)







#encoding:utf-8
import os
import importlib
import warnings
from ..utils.utils import ensure_dir
class WriterTensorboardX():
    def __init__(self, writer_dir, logger, enable):
        self.writer = None
        ensure_dir(writer_dir)
        if enable:
            log_path = writer_dir
            try:
                self.writer = importlib.import_module('tensorboardX').SummaryWriter(log_path)
            except ModuleNotFoundError:
                message = """TensorboardX visualization is configured to use, but currently not installed on this machine. Please install the package by 'pip install tensorboardx' command or turn off the option in the 'configs.json' file."""
                warnings.warn(message, UserWarning)
                logger.warn()
        self.step = 0
        self.mode = ''

        self.tensorboard_writer_ftns = ['add_scalar', 'add_scalars', 'add_image', 'add_audio', 'add_text', 'add_histogram', 'add_pr_curve', 'add_embedding']

    def set_step(self, step, mode='train'):
        self.mode = mode
        self.step = step

    def __getattr__(self, name):
        """
        If visualization is configured to use:
            return add_data() methods of tensorboard with additional information (step, tag) added.
        Otherwise:
            return blank function handle that does nothing
        """
        if name in self.tensorboard_writer_ftns:
            add_data = getattr(self.writer, name, None)
            def wrapper(tag, data, *args, **kwargs):
                if add_data is not None:
                    add_data('{}/{}'.format(self.mode, tag), data, self.step, *args, **kwargs)
            return wrapper
        else:
            # default action for returning methods defined in this class, set_step() for instance.
            try:
                attr = object.__getattr__(name)
            except AttributeError:
                raise AttributeError("type object 'WriterTensorboardX' has no attribute '{}'".format(name))
            return attr


#encoding:utf-8
import numpy as np

class EarlyStopping(object):
    def __init__(self,
                 min_delta=0,
                 patience=10,
                 verbose=1,
                 mode='min',
                 monitor='loss',
                 logger = None,
                 baseline=None):
        '''
        # Arguments
            min_delta: 最小变化
            patience: 多少个epoch未提高，就停止训练
            verbose: 信息大于，默认打印信息
            mode: 计算模式
            monitor: 计算指标
            baseline: 基线
        '''
        self.baseline = baseline
        self.patience = patience
        self.verbose = verbose
        self.min_delta = min_delta
        self.monitor = monitor
        self.logger = logger

        assert mode in ['min','max'],"mode == 'min' or mode == 'max'"
        self.use = 'on_epoch_end'

        if mode == 'min':
            self.monitor_op = np.less
        elif mode == 'max':
            self.monitor_op = np.greater

        if self.monitor_op == np.greater:
            self.min_delta *= 1
        else:
            self.min_delta *= -1
        self._reset()

    def _reset(self):
        # Allow instances to be re-used
        self.wait = 0
        self.stop_training = False
        if self.baseline is not None:
            self.best = self.baseline
        else:
            self.best = np.Inf if self.monitor_op == np.less else -np.Inf

    def step(self,current,epoch):
        if self.monitor_op(current - self.min_delta, self.best):
            self.best = current
            self.wait = 0
        else:
            self.wait += 1
            if self.wait >= self.patience:
                if self.verbose >0:
                    self.logger.info("{patience} epochs with no improvement after which training will be stopped".format(patience = self.patience))
                self.stop_training = True



#encoding:utf-8
import numpy as np
import json
from os import path
from ..utils.utils import ensure_dir
import matplotlib.pyplot as plt
plt.switch_backend('agg') # 防止ssh上绘图问题

class TrainingMonitor():
    def __init__(self, fig_dir, arch,json_dir=None, start_at=0):
        '''
        :param figPath: 保存图像路径
        :param jsonPath: 保存json文件路径
        :param startAt: 重新开始训练的epoch点
        '''
        self.start_at = start_at
        self.H = {}
        self.loss_path = path.sep.join([fig_dir,arch+'_loss.png'])
        self.acc_path = path.sep.join([fig_dir,arch+"_accuracy.png"])
        self.f1_path  = path.sep.join([fig_dir,arch+"_f1.png"])
        self.json_path = path.sep.join([json_dir,arch+"_training_monitor.json"])
        self.use = 'on_epoch_end'

        ensure_dir(fig_dir)
        ensure_dir(json_dir)

    def _restart(self):
        if self.start_at > 0:
            # 如果jsonPath文件存在，咋加载历史训练数据
            if self.json_path is not None:
                if path.exists(self.json_path):
                    self.H = json.loads(open(self.json_path).read())
                    for k in self.H.keys():
                        self.H[k] = self.H[k][:self.start_at]

    def step(self,logs={}):
        for (k, v) in logs.items():
            l = self.H.get(k, [])
            # np.float32会报错
            if not isinstance(v,np.float):
                v = round(float(v),4)
            l.append(v)
            self.H[k] = l

        # 写入文件
        if self.json_path is not None:
            f = open(self.json_path, "w")
            f.write(json.dumps(self.H))
            f.close()

        #保存train图像
        if len(self.H["loss"]) > 1:
            # plot the training loss and accuracy
            N = np.arange(0, len(self.H["loss"]))
            plt.style.use("ggplot")
            plt.figure()
            plt.plot(N, self.H["loss"],label="train_loss")
            plt.plot(N, self.H["val_loss"],label="val_loss")
            plt.legend()
            plt.xlabel("Epoch #")
            plt.ylabel("Loss")
            plt.title("Training Loss [Epoch {}]".format(len(self.H["loss"])))
            plt.savefig(self.loss_path)
            plt.close()

            plt.figure()
            plt.plot(N, self.H["acc"], label="train_acc")
            plt.plot(N, self.H["val_acc"], label="val_acc")
            # plt.plot(N, self.H["rank"], label="train_rank5")
            # plt.plot(N, self.H["val_rank"], label="val_rank5")
            plt.legend()
            plt.xlabel("Epoch #")
            plt.ylabel("Accuracy")
            plt.title("Training Accuracy [Epoch {}]".format(len(self.H["acc"])))
            plt.savefig(self.acc_path)
            plt.close()

            plt.figure()
            plt.plot(N, self.H["f1"], label="train_f1")
            plt.plot(N, self.H["val_f1"], label="val_f1")
            plt.legend()
            plt.xlabel("Epoch #")
            plt.ylabel("F1 Score")
            plt.title("Training F1 Score [Epoch {}]".format(len(self.H["f1"])))
            plt.savefig(self.f1_path)
            plt.close()







#encoding:utf-8
import json
from collections import defaultdict

def format_result(result,words,name):
    '''
    形式调整
    :param result:
    :param words:
    :return:
    '''
    entities = defaultdict(list)
    entities['text'] = "".join(words)
    if len(result) == 0:
        entities['info'] = result
    else:
        for record in result:
            begin = record['begin']
            end  = record['end']
            if (name == 'CWS'):
                record['type'] = 'word'
            entities['info'].append({
                'start':begin+1,
                'end':end+1,
                'word': "".join(words[begin:end + 1]),
                'type':record['type']
            })
    return entities

def get_entity(path,tag_map):
    results = []
    record = {}
    for index,tag_id in enumerate(path):
        
        if tag_id == 0: # 0是我们的pad label
            continue
        tag = tag_map[tag_id]
        if tag.startswith("B_"):
            if record.get('end'):
                if (record['type']=='T'):
                    continue
                results.append(record)
            record = {}
            record['begin'] = index
            record['type'] = tag.split('_')[1]
        elif tag.startswith('I_') and record.get('begin') != None:
            tag_type = tag.split('_')[1]
            if tag_type == record['type']:
                record['end'] = index
        else:
            if record.get('end'):
                if (record['type']=='T'):
                    continue
                results.append(record)
            record = {}
    return results
def get_word(path):

    results = []
    record = {}

    for index,tag_id in enumerate(path):
        
        if tag_id == 0: # 0是我们的pad label
            continue
        if tag_id == 1:
            if (record.get('begin')):
                record['end']=index-1
                results.append(record)
                record={}
            record['begin'] = index
        elif tag_id == 2:
            pass
        else:
            if (record.get('begin')):
                record['end']=index-1
                results.append(record)
                record={}
            results.append({'begin':index,'end':index})

    if (record.get('begin')):
        record['end']=len(path)-1
        results.append(record)
        record={}


    return results

def test_write(data,filename,raw_text_path, name):
    '''
    将test结果保存到文本中，这里是以json格式写入
    :param data:
    :param filename:
    :param raw_text_path:
    :return:
    '''
    with open(raw_text_path,'r') as fr,open(filename,'w',encoding='utf-8') as fw:
        for text,result in zip(fr.readlines(),data):
            words = text.strip()
            words = words.split(' ')
            record = format_result(result,words,name)

            
            encode_json = json.dump(record,fw,ensure_ascii=False)
            print('\n',end='',file=fw)
            #data = encode_json.encode('utf-8').decode('utf-8')
            #print(json.loads(encode_json))
            #print(data, file=fw)



#encoding:utf-8
import os
import random
import operator
import numpy as np
from tqdm import tqdm
from collections import Counter
from ..utils.utils import pkl_read,pkl_write,text_write
from ..utils.gazetteer import Gazetteer


class DataTransformer(object):

    def __init__(self,
                 vocab_path,
                 rev_vocab_path,
                 logger,
                 max_features = None,
                 min_freq = 3,
                 all_data_path = None,
                 label_to_id = None,
                 train_file = None,
                 valid_file = None,
                 test_file = None,
                 valid_size = None,
                 skip_header = False,
                 is_train_mode = True,
                 default_token = False,
                 seed=1024,
                 word_dic = False
                 ):

        self.seed = seed
        self.logger = logger
        self.valid_size = valid_size
        self.min_freq = min_freq
        self.train_file = train_file
        self.valid_file = valid_file
        self.test_file = test_file
        self.all_data_path = all_data_path
        self.vocab_path = vocab_path
        self.rev_vocab_path = rev_vocab_path
        self.word_vocab = None
        self.skip_header = skip_header
        self.max_features = max_features
        self.label_to_id = label_to_id
        self.is_train_mode = is_train_mode
        self.default_token = default_token
        self.word_dic = word_dic


    def _split_sent(self,line):

        res = line.strip('\n').split()
        return res

    def _word_to_id(self,word, vocab):

        return vocab[word] if word in vocab else vocab['<unk>']

    def train_val_split(self,X, y, valid_size=0.3, random_state=2018, shuffle=True):

        self.logger.info('train val split')
        data = []
        for data_x, data_y in tqdm(zip(X, y), desc='Merge'):
            data.append((data_x, data_y))
        del X, y
        N = len(data)
        test_size = int(N * valid_size)
        if shuffle:
            random.seed(random_state)
        random.shuffle(data)
        valid = data[:test_size]
        train = data[test_size:]
        return train,valid

    def build_vocab(self):
        if os.path.isfile(self.vocab_path):
            self.vocab = pkl_read(self.vocab_path)
            self.rev_vocab = pkl_read(self.rev_vocab_path)
        else:
            count = Counter()
            for path in self.all_data_path:
                with open(path, 'r') as fr:
                    self.logger.info('Building word vocab')
                    for i,line in enumerate(fr):

                        if i==0 and self.skip_header:
                            continue
                        words = self._split_sent(line)
                        count.update(words)
            count = {k: v for k, v in count.items()}
            count = sorted(count.items(), key=operator.itemgetter(1))

            all_words = [w[0] for w in count if w[1] >= self.min_freq]
            if self.max_features:
                all_words = all_words[:self.max_features]

            flag_words = ['<pad>', '<unk>']
            all_words = flag_words + all_words
            self.logger.info('vocab_size is %d' % len(all_words))

            word2id = {k: v for k, v in zip(all_words, range(0, len(all_words)))}
            assert word2id['<pad>'] == 0, "ValueError: '<pad>' id is not 0"

            pkl_write(data = word2id,filename=self.vocab_path)
            self.vocab = word2id
            self.rev_vocab = {v : k for k, v in zip(all_words, range(0, len(all_words)))}
            pkl_write(data = self.rev_vocab, filename=self.rev_vocab_path)

    def sentence2id(self,raw_data_path=None,raw_target_path  =None,x_var = None,y_var = None):


        if self.is_train_mode:
            if os.path.isfile(self.train_file) and os.path.isfile(self.valid_file):
                return True
            sentences, labels = [], []
            with open(raw_data_path, 'r') as fr_x,open(raw_target_path,'r') as fr_y:
                for i,(sent,target) in enumerate(zip(fr_x,fr_y)):
                    if i==0 and self.skip_header:
                        continue
                    words = self._split_sent(sent)
                    label = self._split_sent(target)
                    if len(words) ==0 or len(label) ==0:
                        continue
                    if (self.default_token):
                        sent2id = words
                        if len(sent2id)>512:
                            ##print("bad_sent")
                            continue
                    else:
                        sent2id = [self._word_to_id(word=word, vocab=self.vocab) for word in words]
                    label = [self.label_to_id[x] for x in label]
                    sentences.append(sent2id)
                    labels.append(label)

            if self.valid_size:
                train, val = self.train_val_split(X = sentences, y = labels,
                                                  valid_size=self.valid_size,
                                                  random_state=self.seed,
                                                  shuffle=True)
                text_write(self.train_file, train,x_var = x_var,y_var = y_var)
                text_write(self.valid_file, val,x_var = x_var,y_var = y_var)
        else:
            if os.path.isfile(self.test_file):
                print("skip")
                return True
            sentences,labels = [],[]
            with open(raw_data_path, 'r') as fr_x:
                for i,sent in enumerate(fr_x):
                    if i==0 and self.skip_header:
                        continue
                    words = self._split_sent(sent)
                    if len(words) ==0:
                        continue

                    if (self.default_token):
                        sent2id = words
                        if len(sent2id)>512:
                            continue
                    else:
                        sent2id = [self._word_to_id(word=word, vocab=self.vocab) for word in words]

                    label    = [-1 for _ in range(len(sent2id))]
                    sentences.append(sent2id)
                    labels.append(label)
            text_write(self.test_file,zip(sentences,labels),x_var = x_var,y_var = y_var)

    def build_embedding_matrix(self,embedding_path,dict_path, emb_mean = None,emb_std = None):

        embeddings_index = self._load_embedding(embedding_path)
        all_embs = np.stack((embeddings_index.values()))
        if emb_mean is None or emb_std is None:
            emb_mean = all_embs.mean()
            emb_std  = all_embs.std()
        embed_size = all_embs.shape[1]
        nb_words = len(self.vocab)
        gaz = Gazetteer(False)
        embedding_matrix = np.random.normal(emb_mean, emb_std, (nb_words, embed_size))
        for word, id in self.vocab.items():
        
            embedding_vector = embeddings_index.get(word)
            if embedding_vector is not None:
                embedding_matrix[id] = embedding_vector

        if (not self.word_dic):
            return embedding_matrix, None, None
        word_embedding_matrix = None
        if dict_path != None:
            words_embedding_index = self._load_embedding(dict_path)
            words = [k for k in words_embedding_index.keys() if len(k)>1]
            word_num = len(words)
            all_embs = np.stack((embeddings_index.values()))
            if emb_mean is None or emb_std is None:
                emb_mean = all_embs.mean()
                emb_std  = all_embs.std()
            embed_size = all_embs.shape[1]
            word_embedding_matrix = np.random.normal(emb_mean,emb_std,(word_num, embed_size))


            self.word_vocab = {k: v for k, v in zip(words, range(0, len(words)))}
            for word, id in self.word_vocab.items():
                char_list = []
                for char in word:
                    try:
                        char_list.append(self.vocab[char])
                    except:
                        char_list = []
                        break   
                if (char_list == []):
                    continue
                gaz.insert(char_list,id)
                embedding_vector = words_embedding_index.get(word)
                if embedding_vector is not None:
                    word_embedding_matrix[id] = embedding_vector
            
                            
        return embedding_matrix, word_embedding_matrix, gaz

    def _load_embedding(self, embedding_path):

        embeddings_index = {}
        f = open(embedding_path, 'r',errors='ignore',encoding = 'utf8')
        for line in f:
            values = line.rstrip().split(' ')
            if len(values)<10:
                continue
            try:
                word  = values[0]
                coefs = np.asarray(values[1:], dtype='float32')
                embeddings_index[word] = coefs
            except:
                print("Error on ", values[:2])
        f.close()
        return embeddings_index


#encoding:utf-8
import torch
import numpy as np
from sklearn.metrics import f1_score, classification_report
from ..test.predict_utils import get_entity
from collections import Counter

class Accuracy(object):

    def __init__(self,topK):
        super(Accuracy,self).__init__()
        self.topK = topK

    def __call__(self, output, target):
        batch_size = target.size(0)
        _, pred = output.topk(self.topK, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target.view(1, -1).expand_as(pred))
        correct_k = correct[:self.topK].view(-1).float().sum(0)
        result = correct_k / batch_size
        return result

class F1_score(object):
    def __init__(self,num_classes = None):
        self.labels = None
        if num_classes:
            self.labels = [i for i in range(num_classes)]
    def __call__(self,best_path,target):
        y_pred = best_path.contiguous().view(1, -1).squeeze().cpu()
        y_true = target.contiguous().view(1, -1).squeeze().cpu()
        y_true = y_true.numpy()
        y_pred = y_pred.numpy()
        f1 = f1_score(y_true, y_pred, labels=self.labels, average="macro")
        correct = np.sum((y_true==y_pred).astype(int))
        acc = correct/y_pred.shape[0]
        return (acc, f1)

class Classification_Report(object):
    def __init__(self):
        pass
    def __call__(self,y_pred,y_true):
        _, y_pred = torch.max(y_pred.data, 1)
        y_true = y_true.numpy()
        y_pred = y_pred.numpy()
        classify_report = classification_report(y_true, y_pred)
        print('\n\nclassify_report:\n', classify_report)


# 实体得分情况
class Entity_Score(object):
    def __init__(self,id_to_label):
        self.id_to_label = id_to_label
        self._reset()

    def _reset(self):
        self.origins = []
        self.founds = []
        self.rights = []

    def _compute(self,origin,found,right):
        recall =0 if origin ==0 else (right / origin)
        precision= 0 if found == 0 else (right / found)
        f1 = 0. if recall + precision ==0 else (2 * precision* recall) / (precision + recall)
        return recall,precision,f1

    def result(self):

        origin_counter = Counter([x['type'] for x in self.origins])
        found_counter = Counter([x['type'] for x in self.founds])
        right_counter = Counter([x['type'] for x in self.rights])
        for type,count in origin_counter.items():

            origin = count
            found = found_counter.get(type,0)
            right = right_counter.get(type,0)
            recall,precision,f1 = self._compute(origin,found,right)
            print("count=",count)
            print("Type: %s - precision: %.4f - recall: %.4f - f1: %.4f"%(type,recall,precision,f1))

    def update(self,label_paths,pred_paths):
        '''
        :param label_paths: [[2,3,4,5,6,5,8,8,8,8,8,4],[2,3,7,7,7,8,8,8,8]]
        :param pred_paths: [[2,3,4,5,6,5,8,8,8,8,8,4],[2,3,7,7,7,8,8,8,8]]
        :return:
        '''

        for label_path,pre_path in zip(label_paths,pred_paths):
            label_entities = get_entity(label_path,self.id_to_label)
            pre_entities = get_entity(pre_path, self.id_to_label)
            self.origins.extend(label_entities)
            self.founds.extend(pre_entities)
            self.rights.extend([pre_entity for pre_entity in pre_entities if pre_entity in label_entities])







#encoding:utf-8
import os
import sys
import torch
import torch.nn as nn

# gpu
def prepare_device(n_gpu_use,logger):
    """
    setup GPU device if available, move model into configured device
    # 如果n_gpu_use为数字，则使用range生成list
    # 如果输入的是一个list，则默认使用list[0]作为controller
     """
    if isinstance(n_gpu_use,int):
        n_gpu_use = range(n_gpu_use)
    n_gpu = torch.cuda.device_count()
    if len(n_gpu_use) > 0 and n_gpu == 0:
        logger.warning("Warning: There\'s no GPU available on this machine, training will be performed on CPU.")
        n_gpu_use = range(0)
    if len(n_gpu_use) > n_gpu:
        msg = "Warning: The number of GPU\'s configured to use is {}, but only {} are available on this machine.".format(n_gpu_use, n_gpu)
        logger.warning(msg)
        n_gpu_use = range(n_gpu)
    device = torch.device('cuda:%d'%n_gpu_use[0] if len(n_gpu_use) > 0 else 'cpu')
    list_ids = n_gpu_use
    return device, list_ids

# 加载模型
def restore_checkpoint(resume_path,model = None,optimizer = None):
    checkpoint = torch.load(resume_path)
    best = checkpoint['best']
    start_epoch = checkpoint['epoch'] + 1
    if model:
        model.module.load_state_dict(checkpoint['state_dict'])

    if optimizer:
        optimizer.load_state_dict(checkpoint['optimizer'])

    return [model,optimizer,best,start_epoch]

# 判断环境 cpu还是gpu
def model_device(n_gpu,model,logger):
    device, device_ids = prepare_device(n_gpu,logger)

    os.environ['CUDA_VISIBLE_DEVICES'] = str(device_ids[0])
    #model = model.to(device)
    return model,device

# 对batch数据
def batchify_with_label(inputs,outputs,target = None,is_train_mode =True):
    # batch_size * seq_len * num_classes

    mask = inputs >0
    mask = mask[:,:outputs.size(1)]
    if is_train_mode:
        target = target[:,:outputs.size(1)]

    return mask,target


#encoding:utf-8
import os
import numpy as np
import torch
from ..utils.utils import ensure_dir

class ModelCheckpoint(object):

    def __init__(self, checkpoint_dir,
                 monitor,
                 logger,
                 arch,
                 save_best_only = True,
                 best_model_name = None,
                 epoch_model_name = None,
                 mode='min',
                 epoch_freq=1,
                 best = None):
        self.monitor = monitor
        self.checkpoint_dir = checkpoint_dir
        self.save_best_only = save_best_only
        self.epoch_freq = epoch_freq
        self.arch = arch
        self.logger = logger
        self.best_model_name = best_model_name
        self.epoch_model_name = epoch_model_name
        self.use = 'on_epoch_end'

        # 计算模式
        if mode == 'min':
            self.monitor_op = np.less
            self.best = np.Inf

        elif mode == 'max':
            self.monitor_op = np.greater
            self.best = -np.Inf
        # 这里主要重新加载模型时候
        #对best重新赋值
        if best:
            self.best = best
        ensure_dir(self.checkpoint_dir.format(arch = self.arch))

    def step(self, state,current):
        # 是否保存最好模型
        if self.save_best_only:
            if self.monitor_op(current, self.best):
                self.logger.info('\nEpoch %d: %s improved from %0.5f to %0.5f'% (state['epoch'], self.monitor, self.best,current))
                self.best = current
                state['best'] = self.best
                best_path = os.path.join(self.checkpoint_dir.format(arch = self.arch), self.best_model_name.format(arch = self.arch))
                torch.save(state, best_path)
        # 每隔几个epoch保存下模型
        else:
            filename = os.path.join(self.checkpoint_dir.format(arch = self.arch), self.epoch_model_name.format(arch=self.arch,
                                                                                epoch=state['epoch'],
                                                                                val_loss=state[self.monitor]
                                                                                )
                                    )
            if state['epoch'] % self.epoch_freq == 0:
                self.logger.info("\nEpoch %d: save model to disk."%(state['epoch']))
                torch.save(state, filename)
